import uuid
from .solution_component import SolutionComponent
from .paged_query_result import PagedQueryResult


class ApiInterface:
    """
    Interface with all methods available both to the pipeline runners and the Data Science API.
    """

    def __init__(self):
        self._dto_registry = {}

    def get(self,
            obj=None,
            id: uuid.UUID = None,
            key: str = None,
            entity=None):
        """
        get record content from DS API.

        - get will look first for obj then for id then for key (e.g. if id and key specified, key is ignored )

        :param obj: a wizata_dsapi.<Entity>() with UUID set - fetch using technical UUID

        :param id: a UUID of a specific object (use in combination of entity)
        :param key: a logical key of a specific object (e.g. hardware id for DataPoint and Twin (use in combination of entity)
        :param entity: plural name of the entity or class (use in combination of entity)(e.g. scripts, plots, mlmodels, dataframes...)
        :return: object with all properties or None if not found.
        """
        pass

    def _get_class(self,
                   entity,
                   operation: str = None):
        """
        get ApiDto class definition for a specific entity and valid that desired operation is supported.
        :param entity: str plural name of the entity or direct class definition.
        :param operation: optional - set operation name to verify it is valid on the class.
        """
        if isinstance(entity, str):
            if entity in self._dto_registry.keys():
                if operation is not None and operation not in self._dto_registry[entity].operations():
                    raise NotImplementedError(f"Operation {operation} not supported for {entity} entity.")
                return self._dto_registry[entity]
            else:
                raise TypeError(f"Api entity {entity} is not supported.")
        elif entity in self._dto_registry.values():
            if operation is not None and operation not in entity.operations():
                raise NotImplementedError(f"Operation {operation} not supported for {entity} entity.")
            return entity
        else:
            raise TypeError(f"Api entity {entity} is not supported.")

    def get_components(self,
                       label_id: uuid.UUID = None,
                       twin_id: uuid.UUID = None,
                       template_id: uuid.UUID = None,
                       owner_id: uuid.UUID = None,
                       organization_only: bool = False,
                       name: str = None):
        """
        get components
        :param label_id: filter on a specific label
        :param template_id: filter on a specific template
        :param twin_id: filter on a specific twin
        :param owner_id: filter on a specific owner_id
        :param organization_only: work only with organization components (by default - False)
        :param name: filter on a specific name (contains)
        """
        pass

    def create_component(self, component: SolutionComponent):
        """
        create a component based on its ID.
        """
        pass

    def update_component(self, component: SolutionComponent):
        """
        update a component based on its ID.
        """
        pass

    def delete_component(self, component_id: uuid.UUID):
        """
        delete component
        """
        pass

    def get_business_labels(self) -> dict:
        """
        get a name / uuid dictionary with all business labels in platform.
        """
        pass

    def get_datapoint_mappings(self, registration):
        """
        get datapoint mapping from a registration.
        """
        pass

    def get_registrations(self, template) -> list:
        """
        retrieve all registrations for
        :param template: template object, UUID or str key.
        :return: list of twin registration.
        """
        pass

    def search_datapoints(self,
                          page: int = 1,
                          size: int = 20,
                          sort: str = "id",
                          direction: str = "asc",
                          hardware_id: str = None,
                          categories: list = None,
                          business_types: list = None,
                          twin=None,
                          recursive: bool = False) -> PagedQueryResult:
        """
        get datapoints with a paged query.
        :param page: numero of the page - default 1.
        :param size: quantity per page - default 20 max 100.
        :param sort: column to sort results - default id.
        :param direction: sorting direction by default asc, accept also desc.
        :param hardware_id: filter on a specific hardware ID name or partial name.
        :param business_types: list of BusinessType or str.
        :param categories: list of UUID or Category.
        :param twin: uuid or Twin element to search datapoints.
        :param recursive: set to True in combination of a twin to look inside all sub-twins recursively.
        :return: PagedQueryResults, check total for number of potential results and results for the list of entity.
        """
        pass

    def search_twins(self,
                     page: int = 1,
                     size: int = 20,
                     sort: str = "id",
                     direction: str = "asc",
                     hardware_id: str = None,
                     name: str = None,
                     parents: list = None) -> PagedQueryResult:
        """
        get twins with a paged query.
        :param page: numero of the page - default 1.
        :param size: quantity per page - default 20 max 100.
        :param sort: column to sort results - default id.
        :param direction: sorting direction by default asc, accept also desc.
        :param hardware_id: filter on a specific hardware ID name or partial name.
        :param name: name or part of twin name.
        :param parents: list of all possible parents (Twin, UUID, or str UUID).
        :return: PagedQueryResults, check total for number of potential results and results for the list of entity.
        """
        pass



import uuid
from .api_dto import ApiDto


class BusinessLabel(ApiDto):

    @classmethod
    def route(cls):
        return "components/labels"

    @classmethod
    def from_dict(cls, data):
        obj = BusinessLabel()
        obj.from_json(data)
        return obj

    @classmethod
    def operations(cls):
        return ['lists']

    """
    A Business Label is a tag representing a business concept.
    """
    def __init__(self, label_id=None, name=None, description=None, icon=None, color=None, order=None):
        if label_id is None:
            self.label_id = uuid.uuid4()
        else:
            self.label_id = label_id
        self.name = name
        self.description = description
        self.icon = icon
        self.color = color
        self.order = order
        self.createdById = None
        self.createdDate = None
        self.updatedById = None
        self.updatedDate = None

    def api_id(self) -> str:
        return str(self.label_id).upper()

    def endpoint(self) -> str:
        return "BusinessLabels"

    def to_json(self, target: str = None):
        obj = {
            "id": str(self.label_id)
        }
        if self.name is not None:
            obj["name"] = str(self.name)
        if self.description is not None:
            obj["description"] = str(self.description)
        if self.icon is not None:
            obj["icon"] = str(self.icon)
        if self.color is not None:
            obj["color"] = str(self.color)
        if self.order is not None:
            obj["order"] = int(self.order)
        if self.createdById is not None:
            obj["createdById"] = str(self.createdById)
        if self.createdDate is not None:
            obj["createdDate"] = str(self.createdDate)
        if self.updatedById is not None:
            obj["updatedById"] = str(self.updatedById)
        if self.updatedDate is not None:
            obj["updatedDate"] = str(self.updatedDate)
        return obj

    def from_json(self, obj):
        if "id" in obj.keys():
            self.label_id = uuid.UUID(obj["id"])
        if "name" in obj.keys() and obj["name"] is not None:
            self.name = obj["name"]
        if "description" in obj.keys() and obj["description"] is not None:
            self.description = obj["description"]
        if "icon" in obj.keys() and obj["icon"] is not None:
            self.icon = obj["icon"]
        if "color" in obj.keys() and obj["color"] is not None:
            self.color = obj["color"]
        if "order" in obj.keys() and obj["order"] is not None:
            self.order = obj["order"]
        if "createdById" in obj.keys() and obj["createdById"] is not None:
            self.createdById = obj["createdById"]
        if "createdDate" in obj.keys() and obj["createdDate"] is not None:
            self.createdDate = obj["createdDate"]
        if "updatedById" in obj.keys() and obj["updatedById"] is not None:
            self.updatedById = obj["updatedById"]
        if "updatedDate" in obj.keys() and obj["updatedDate"] is not None:
            self.updatedDate = obj["updatedDate"]



import uuid
from .api_dto import ApiDto
from enum import Enum
import json


class SolutionType(Enum):
    GRAFANA = "Grafana"
    DASHBOARD = "Dashboard"
    IFRAME = "Iframe"


class SolutionComponent(ApiDto):
    """
    A solution components handle an element giving to end-users to interact with solution.
    :ivar label_id: Business Label identification to set the menu and category.
    :ivar dashboard_id: Used only for type DASHBOARD as link to the dashboard.
    """

    def __init__(self,
                 solution_component_id=None,
                 label_id=None,
                 name=None,
                 solution_type=None,
                 content=None,
                 order=None,
                 dashboard_id=None,
                 twin_id=None,
                 template_id=None,
                 owner_id=None):
        if solution_component_id is None:
            self.solution_component_id = uuid.uuid4()
        else:
            self.solution_component_id = solution_component_id

        self.label_id = label_id
        self.name = name

        self.order = order
        if self.order is not None and not isinstance(self.order, int):
            raise TypeError(f'order must be None or a valid integer')

        self.solution_type = solution_type
        if self.solution_type is not None and not isinstance(self.solution_type, SolutionType):
            raise TypeError(f'solution type must be None or a valid SolutionType')

        self.content = content
        self.dashboard_id = dashboard_id

        self.owner_id = owner_id

        self.twin_id = twin_id
        self.template_id = template_id

        self.createdById = None
        self.createdDate = None
        self.updatedById = None
        self.updatedDate = None

    def api_id(self) -> str:
        return str(self.solution_component_id).upper()

    def endpoint(self) -> str:
        return "Components"

    def to_json(self, target: str = None):
        if self.label_id is None:
            raise ValueError("label ID is required when creating/updating a component.")
        if self.name is None:
            raise ValueError("name is required when creating/updating a component.")
        if self.solution_type is None or not isinstance(self.solution_type, SolutionType):
            raise ValueError("solution type must be a valid solution type.")
        obj = {
            "id": str(self.solution_component_id),
            "labelId": str(self.label_id),
            "name": self.name,
            "type": self.solution_type.value
        }

        if self.order is not None:
            if not isinstance(self.order, int):
                raise ValueError('order must be None or a valid integer.')
            obj["order"] = self.order

        if self.content is not None:
            obj["content"] = self.content
        if self.solution_type == SolutionType.DASHBOARD:
            if self.dashboard_id is None or not isinstance(self.dashboard_id, uuid.UUID):
                raise ValueError("dashboard Id of type UUID must be on component type dashboard")
            obj["dashboardId"] = self.dashboard_id

        if self.owner_id is not None:
            obj["ownerId"] = str(self.owner_id)
        if self.twin_id is not None:
            obj["twinId"] = str(self.twin_id)
        if self.template_id is not None:
            obj["templateId"] = str(self.template_id)

        return obj

    def from_json(self, obj):
        if "id" in obj.keys():
            self.solution_component_id = uuid.UUID(obj["id"])
        if "labelId" in obj.keys() and obj["labelId"] is not None:
            self.label_id = uuid.UUID(obj["labelId"])
        if "name" in obj.keys() and obj["name"] is not None:
            self.name = obj["name"]
        if "ownerId" in obj.keys() and obj["ownerId"] is not None:
            self.owner_id = uuid.UUID(obj["ownerId"])
        if "order" in obj.keys() and obj["order"] is not None:
            self.order = int(obj["order"])
        if "content" in obj.keys() and obj["content"] is not None:
            self.content = obj["content"]
        if "type" in obj.keys():
            self.solution_type = SolutionType(str(obj["type"]))
        if "twinId" in obj.keys() and obj["twinId"] is not None:
            self.twin_id = uuid.UUID(obj["twinId"])
        if "dashboardId" in obj.keys() and obj["dashboardId"] is not None:
            self.dashboard_id = uuid.UUID(obj["dashboardId"])
        if "templateId" in obj.keys() and obj["templateId"] is not None:
            self.template_id = uuid.UUID(obj["templateId"])
        if "createdById" in obj.keys() and obj["createdById"] is not None:
            self.createdById = obj["createdById"]
        if "createdDate" in obj.keys() and obj["createdDate"] is not None:
            self.createdDate = obj["createdDate"]
        if "updatedById" in obj.keys() and obj["updatedById"] is not None:
            self.updatedById = obj["updatedById"]
        if "updatedDate" in obj.keys() and obj["updatedDate"] is not None:
            self.updatedDate = obj["updatedDate"]



from .execution import Execution, ExecutionStepLog


class ILogger:
    """
    logger interface
    """

    def write_log(self, message: str, level: int = 7):
        pass

    def notify_step(self, step_log: ExecutionStepLog):
        pass

    def notify_execution(self, execution: Execution):
        pass



import uuid
import pandas
from .api_dto import ApiDto


class DSDataFrame(ApiDto):
    """
    A DS Dataframe is a definition of a pandas Dataframe that can be stored and shared on Wizata.

    :ivar df_id: The UUID of the dataframe.
    :ivar generatedById: The UUID of the Execution from which the dataframe was created.
    :ivar dataframe: Pandas Dataframe.
    """

    @classmethod
    def route(cls):
        return "dataframes"

    @classmethod
    def from_dict(cls, data):
        obj = DSDataFrame()
        obj.from_json(data)
        return obj

    @classmethod
    def operations(cls):
        return ['lists', 'get_by_id', 'get_by_key', 'delete']

    @classmethod
    def get_type(cls):
        return "pickle"

    def __init__(self, df_id=None, dataframe=None):
        if df_id is None:
            self.df_id = uuid.uuid4()
        else:
            self.df_id = df_id
        self.generatedById = None
        self.dataframe = dataframe

    def api_id(self) -> str:
        """
        Id of the dataframe (df_id)

        :return: string formatted UUID of the dataframe.
        """
        return str(self.df_id).upper()

    def endpoint(self) -> str:
        """
        Name of the endpoints used to manipulate dataframes.
        :return: Endpoint name.
        """
        return "Dataframes"

    def from_json(self, obj):
        """
        Load the Dataframe entity from a dictionary representation of the dataframe.
        Doesn't load the dataframe itself, you can use df_from_json to fetch it from JSON, CSV or other formats.

        :param obj: Dict version of the Dataframe.
        """
        if "id" in obj.keys():
            self.df_id = uuid.UUID(obj["id"])
        if "generatedById" in obj.keys():
            self.generatedById = obj["generatedById"]

    def to_json(self, target: str = None):
        """
        Convert the dataframe to a dictionary compatible to JSON format.

        :return: dictionary representation of the Dataframe object.
        """
        obj = {
            "id": str(self.df_id)
        }
        if self.generatedById is not None:
            obj["generatedById"] = str(self.generatedById)
        return obj




import uuid
import json
from .api_dto import ApiDto
from .pipeline import VarType
from .datapoint import DataPoint
from .template import TemplateProperty


class TwinRegistrationProperty(ApiDto):

    def __init__(self,
                 twin_registration_property_id: uuid.UUID = None,
                 twin_registration_id: uuid.UUID = None,
                 template_property: TemplateProperty = None,
                 value=None,
                 p_type: VarType = None):
        if twin_registration_property_id is None:
            self.twin_registration_property_id = uuid.uuid4()
        else:
            self.twin_registration_property_id = twin_registration_property_id
        self.twin_registration_id = twin_registration_id
        self.template_property = template_property
        self.value = value
        self.p_type = p_type
        self.createdById = None
        self.createdDate = None
        self.updatedById = None
        self.updatedDate = None

    def api_id(self) -> str:
        """
        Id
        :return: string formatted UUID of the template.
        """
        return str(self.twin_registration_property_id).upper()

    def endpoint(self) -> str:
        """
        Name of the endpoints.
        :return: Endpoint name.
        """
        return "TwinRegistrationProperties"

    def from_json(self, obj):
        """
        Load the property from a dictionary.
        :param obj: Dict version of the Property.
        """
        if "id" in obj.keys():
            self.twin_registration_property_id = uuid.UUID(obj["id"])
        if "templatePropertyId" in obj.keys() and obj["templatePropertyId"] is not None:
            self.template_property = TemplateProperty(obj["templatePropertyId"])
        if "twinRegistrationId" in obj.keys() and obj["twinRegistrationId"] is not None:
            self.twin_registration_id = obj["twinRegistrationId"]

        if "createdById" in obj.keys() and obj["createdById"] is not None:
            self.createdById = obj["createdById"]
        if "createdDate" in obj.keys() and obj["createdDate"] is not None:
            self.createdDate = obj["createdDate"]
        if "updatedById" in obj.keys() and obj["updatedById"] is not None:
            self.updatedById = obj["updatedById"]
        if "updatedDate" in obj.keys() and obj["updatedDate"] is not None:
            self.updatedDate = obj["updatedDate"]

        if "sensorId" in obj.keys() and obj["sensorId"] is not None:
            self.value = DataPoint(datapoint_id=obj["sensorId"])
            self.p_type = VarType.DATAPOINT
        elif "json" in obj.keys() and obj["json"] is not None:
            if isinstance(obj["json"], str):
                self.value = json.loads(obj["json"])
            else:
                self.value = obj["json"]
            self.p_type = VarType.JSON
        elif "relative" in obj.keys() and obj["relative"] is not None:
            self.value = obj["relative"]
            self.p_type = VarType.RELATIVE
        elif "datetime" in obj.keys() and obj["datetime"] is not None:
            self.value = obj["datetime"]
            self.p_type = VarType.DATETIME
        elif "integer" in obj.keys() and obj["integer"] is not None:
            self.value = obj["integer"]
            self.p_type = VarType.INTEGER
        elif "float" in obj.keys() and obj["float"] is not None:
            self.value = obj["float"]
            self.p_type = VarType.FLOAT
        elif "string" in obj.keys() and obj["string"] is not None:
            self.value = obj["string"]
            self.p_type = VarType.STRING

    def to_json(self, target: str = None):
        """
        Convert the registration to a dictionary compatible to JSON format.

        :return: dictionary representation of the Registration object.
        """
        obj = {
            "id": str(self.twin_registration_property_id)
        }
        if self.template_property is not None:
            obj["templatePropertyId"] = str(self.template_property.template_property_id)
        if self.twin_registration_id is not None:
            obj["twinRegistrationId"] = str(self.twin_registration_id)
        if self.p_type is not None:
            if self.p_type == VarType.JSON:
                if isinstance(self.value, str):
                    obj["json"] = self.value
                else:
                    obj["json"] = json.dumps(self.value)
            elif self.p_type == VarType.FLOAT:
                try:
                    obj["float"] = float(self.value)
                except Exception as e:
                    print(f'{self.twin_registration_property_id} have invalid property type/value')
                    obj["float"] = None
            elif self.p_type == VarType.STRING:
                obj["string"] = str(self.value)
            elif self.p_type == VarType.INTEGER:
                try:
                    obj["integer"] = int(self.value)
                except Exception as e:
                    print(f'{self.twin_registration_property_id} have invalid property type/value')
                    obj["integer"] = None
            elif self.p_type == VarType.DATETIME:
                try:
                    obj["datetime"] = int(self.value)
                except Exception as e:
                    print(f'{self.twin_registration_property_id} have invalid property type/value')
                    obj["datetime"] = None
            elif self.p_type == VarType.RELATIVE:
                obj["relative"] = str(self.value)
            elif self.p_type == VarType.DATAPOINT:
                obj["sensorId"] = str(self.value.datapoint_id)
            else:
                raise ValueError(f'unrecognized {self.p_type} registration property type')

        if self.createdById is not None:
            obj["createdById"] = str(self.createdById)
        if self.createdDate is not None:
            obj["createdDate"] = str(self.createdDate)
        if self.updatedById is not None:
            obj["updatedById"] = str(self.updatedById)
        if self.updatedDate is not None:
            obj["updatedDate"] = str(self.updatedDate)
        return obj


class TwinRegistration(ApiDto):
    """
    Registration of a Digital Twin on a solution Template.

    :ivar twin_registration_id: UUID of the registration
    :ivar twin_id: UUID of registered Digital Twin
    :ivar template_id: UUID of the solution template.
    :ivar properties: list of Properties { name , datapoint, float, integer, string }

    A property must contains a name and the hardwareId of the datapoint or the value corresponding to the right type.
    """

    def __init__(self, twin_registration_id=None, twin_id=None, template_id=None, properties=None):
        if twin_registration_id is None:
            self.twin_registration_id = uuid.uuid4()
        else:
            self.twin_registration_id = twin_registration_id
        self.twin_id = twin_id
        self.template_id = template_id

        # retro-compatibility properties
        if properties is None:
            properties = []
        self.properties = properties

        # new hidden properties
        self._properties = []

        self.createdById = None
        self.createdDate = None
        self.updatedById = None
        self.updatedDate = None

    def api_id(self) -> str:
        """
        Id of the TwinRegistrations (twin_registration_id)

        :return: string formatted UUID of the template.
        """
        return str(self.twin_registration_id).upper()

    def endpoint(self) -> str:
        """
        Name of the endpoints used to manipulate templates.
        :return: Endpoint name.
        """
        return "TwinRegistration"

    def from_json(self, obj):
        """
        Load the Registration entity from a dictionary.

        :param obj: Dict version of the Registration.
        """
        if "id" in obj.keys():
            self.twin_registration_id = uuid.UUID(obj["id"])
        if "twinId" in obj.keys() and obj["twinId"] is not None:
            self.twin_id = obj["twinId"]
        if "templateId" in obj.keys() and obj["templateId"] is not None:
            self.template_id = obj["templateId"]
        if "properties" in obj.keys() and obj["properties"] is not None:
            if isinstance(obj["properties"], str):
                self.properties = json.loads(obj["properties"])
            else:
                self.properties = obj["properties"]
            self._properties = []
            for dict_property in self.properties:
                if isinstance(dict_property, dict):
                    r_property = TwinRegistrationProperty()
                    r_property.from_json(dict_property)
                    self._properties.append(r_property)
                else:
                    print(f'{dict_property} is invalid properties on registration {self.twin_registration_id} '
                          f'and have been skipped')
        if "createdById" in obj.keys() and obj["createdById"] is not None:
            self.createdById = obj["createdById"]
        if "createdDate" in obj.keys() and obj["createdDate"] is not None:
            self.createdDate = obj["createdDate"]
        if "updatedById" in obj.keys() and obj["updatedById"] is not None:
            self.updatedById = obj["updatedById"]
        if "updatedDate" in obj.keys() and obj["updatedDate"] is not None:
            self.updatedDate = obj["updatedDate"]

    def to_json(self, target: str = None):
        """
        Convert the registration to a dictionary compatible to JSON format.

        :return: dictionary representation of the Registration object.
        """
        obj = {
            "id": str(self.twin_registration_id)
        }
        if self.twin_id is not None:
            obj["twinId"] = str(self.twin_id)
        if self.template_id is not None:
            obj["templateId"] = str(self.template_id)
        if self._properties is not None and self._properties != []:
            list_properties = []
            r_property: TwinRegistrationProperty
            for r_property in self._properties:
                list_properties.append(r_property.to_json())
            obj["properties"] = json.dumps(list_properties)
        elif self.properties is not None:
            obj["properties"] = json.dumps(self.properties)
        if self.createdById is not None:
            obj["createdById"] = str(self.createdById)
        if self.createdDate is not None:
            obj["createdDate"] = str(self.createdDate)
        if self.updatedById is not None:
            obj["updatedById"] = str(self.updatedById)
        if self.updatedDate is not None:
            obj["updatedDate"] = str(self.updatedDate)
        return obj

    def append(self, registration_property: TwinRegistrationProperty):
        """
        append a twin registration property.
        """

        if registration_property.twin_registration_id is None:
            registration_property.twin_registration_id = self.twin_registration_id
        elif registration_property.twin_registration_id != self.twin_registration_id:
            raise ValueError('mismatch between registration id and property id')

        self._properties.append(registration_property)

    def get_properties(self) -> list:
        """
        return a list of TwinRegistrationProperty (replacing dict properties)
        """
        return self._properties

    def get_value(self, name: str, p_type: str):
        """
        get value of a property based on name and type.
        :param name: name of property.
        :param p_type: type of property.
        :return: value.
        """
        if self.properties is None:
            raise ValueError('there is no property on your registration.')

        if name is None or p_type is None:
            raise ValueError('please set a name or a p_type')

        for r_property in self.properties:
            if r_property is None:
                raise ValueError('a registration property cannot be None.')

            if "name" in r_property and r_property["name"] is not None and r_property["name"] == name:
                if p_type not in r_property:
                    raise ValueError(f'requested type is not set on the property {name}')
                return r_property[p_type]

        raise ValueError(f'property {name} of type {p_type} not found in the registration')



import uuid
from datetime import datetime, timedelta, timezone
from .dataframe_toolkit import generate_epoch
from .api_dto import ApiDto

filter_map = {
    '>':'gt',
    '<':'lt',
    '>=':'gte',
    '<=':'lte',
    '==':'eq',
    '!=':'neq'
}


class Request(ApiDto):
    """
    A Request to fetch dataframes from Wizata App.

    To execute the request please use a DS API client.

    :ivar equipments: List of Equipments containing Digital Twin item and datapoints to fetch.

    :ivar template_id: UUID of the template.
    :ivar template: str unique id of the template.
    :ivar twin_id: UUID of the Digital Twin.
    :ivar twin: str hardware id of the Digital Twin.

    :ivar start: Datetime defining beginning of period.
    :ivar end: Datetime defining end of period.
    :ivar aggregation: Aggregation method to fetch, accept "mean" and "stddev".
    :ivar interval: Interval in seconds between each aggregation.
    :ivar filters: Filter to apply on the query (not yet fully implemented).
    :ivar connections: Connections rules between equipment to align data in time (not yet fully implemented).
    :ivar null: By default at 'drop' and dropping NaN values. If not intended behavior please set it to 'ignore' or 'all'.
    """

    @classmethod
    def route(cls):
        return "data"

    @classmethod
    def from_dict(cls, data):
        obj = Request()
        obj.from_json(data)
        return obj

    @classmethod
    def operations(cls):
        return ['query']

    @classmethod
    def get_type(cls):
        return "pickle"

    def __init__(self,
                 datapoints=None,
                 start=None,
                 end=None,
                 agg_method='mean',
                 interval=None,
                 null='drop',
                 template_id=None,
                 template=None,
                 twin_id=None,
                 twin=None,
                 request_id=None,
                 filters=None,
                 options=None):
        if request_id is None:
            request_id = uuid.uuid4()
        self.request_id = request_id
        self.function = None

        # Equipments & Data Points
        self.equipments = []
        if datapoints is not None:
            self.add_datapoints(datapoints)

        # Template & Registration
        self.template = None
        self.select_template(
             template_id=template_id,
             template_key=template,
             twin_id=twin_id,
             twin_hardware_id=twin
        )

        self.start = start
        self.end = end

        self.aggregation = agg_method
        if interval is not None:
            self.interval = int(interval) / 1000

        self.filters = filters
        self.options = options

        self.on_off_sensor = None
        self.restart_time = None
        self.sensitivity = None
        self.dataframe = None
        self.extra_data = None
        self.target_feat = None
        self.connections = None
        self.name = None

        self.null = null

    def prepare(self):
        """
        prepare a dict JSON compatible only for the QUERY part of the request.

        :return: a dict JSON compatible
        """
        query = {}
        if self.request_id is not None:
            query["id"] = str(self.request_id)
        if self.equipments is not None:
            query["equipments_list"] = self.equipments
        else:
            raise KeyError("Missing data points inside the query - add_datapoints")
        if self.start is not None and self.end is not None:

            if isinstance(self.start, str):
                start = self.start
            else:
                start = self.__format_date(self.start)

            if isinstance(self.end, str):
                end = self.end
            else:
                end = self.__format_date(self.end)

            query["timeframe"] = {
                "start": start,
                "end": end
            }
        else:
            raise KeyError("Missing start and end date, please use datatime format")
        if self.aggregation is not None and self.interval:
            query["aggregations"] = {
                "agg_method": self.aggregation,
                "interval": self.interval * 1000
            }
        else:
            raise KeyError("Missing aggregations information inside the request")
        if self.null is not None and self.null != 'drop':
            query['null'] = self.null
        if self.template is not None:
            query['template'] = self.template
            if isinstance(self.template, dict) and 'template_id' in query['template']:
                query['template']['template_id'] = str(query['template']['template_id'])
        if self.filters is not None:
            query['filters'] = self.filters
        if self.options is not None:
            query['options'] = self.options
        return query

    def __format_date(self, dt_to_format):
        if isinstance(dt_to_format, datetime):
            millisec = dt_to_format.timestamp() * 1000
            return int(millisec)
        else:
            raise TypeError("date is not a valid datetime")

    def start_time(self, now=None) -> datetime:
        """
        get start time
        :param now: override now value for relative datetime
        :return: start datetime
        """
        if self.start is None:
            raise ValueError('missing start datetime')
        elif isinstance(self.start, str):
            return datetime.fromtimestamp(generate_epoch(self.start, now=now) / 1000, timezone.utc)
        elif isinstance(self.start, datetime):
            return self.start
        else:
            raise TypeError(f'unsupported start datetime type {self.start.__class__.__name__}')

    def end_time(self, now=None) -> datetime:
        """
        get end time
        :param now: override now value for relative datetime
        :return: start datetime
        """
        if self.end is None:
            raise ValueError('missing end datetime')
        elif isinstance(self.end, str):
            return datetime.fromtimestamp(generate_epoch(self.end, now=now) / 1000, timezone.utc)
        elif isinstance(self.end, datetime):
            return self.end
        else:
            raise TypeError(f'unsupported end datetime type {self.end.__class__.__name__}')

    def add_datapoints(self, datapoints, shift: int = 0):
        """
        Add datapoints to fetch without defining its equipments.
        :param datapoints: List(str) of datapoints to fetch identified by Hardware ID.
        :param shift: Shift to apply in seconds on timestamp, by default 0.
        """
        self.equipments.append({
            "id": None,
            "datapoints": list(datapoints),
            "shift": str(shift) + "s"
        })

    def get_datapoints(self) -> list:
        """
        return list of declared datapoints.
        """
        datapoints = []

        if self.equipments is not None:
            for equipment in self.equipments:
                if "datapoints" not in equipment.keys():
                    raise KeyError("No 'data points' have been provided for equipment with id '" +
                                   str(equipment["id"]) + "'")
                for datapoint in equipment["datapoints"]:
                    if isinstance(datapoint, str):
                        datapoints.append(datapoint)
                    elif "id" in datapoint.keys():
                        datapoints.append(datapoint["id"])
                    else:
                        raise KeyError("Incorrect datapoint declaration : '" + str(datapoint) + "'")

        return datapoints

    def set_datapoints(self, datapoints: list):
        """
        replace current datapoints by provided list
        """
        self.equipments = []
        if datapoints is not None:
            self.add_datapoints(datapoints)

    def add_equipment(self, equipment_id: uuid.UUID, datapoints, shift=0):
        """
        Add datapoints to fetch with a Digital Twin ID identification.
        :param equipment_id: UUID of the Digital Twin ID to which the datapoints are linked.
        :param datapoints: List(str) of datapoints to fetch identified by Hardware ID.
        :param shift: Shift to apply in seconds on timestamp, by default 0.
        """
        if not isinstance(equipment_id, uuid.UUID):
            raise TypeError("equipment_id must be of type uuid.UUID")
        for equipment in self.equipments:
            if "id" in equipment.keys() and equipment["id"] == str(equipment_id):
                raise ValueError("equipment_id is already in the request please remove it before adding datapoints.")
        self.equipments.append({
            "id": str(equipment_id),
            "datapoints": list(datapoints),
            "shift": str(shift) + "s"
        })

    # attempt to remove equipment from the query if exists
    def remove_equipment(self, equipment_id: uuid.UUID):
        """
        Remove an equipment from the list including all its listed datapoints.
        :param equipment_id: UUID of the Digital Twin item.
        """
        if equipment_id is not None and not isinstance(equipment_id, uuid.UUID):
            raise TypeError("equipment_id must be None or of type uuid.UUID")
        found = None
        for equipment in self.equipments:
            if "id" in equipment.keys() and equipment["id"] == str(equipment_id):
                found = equipment
        if found is not None:
            self.equipments.remove(equipment)

    def set_aggregation(self, method, interval):
        """
        Specifies aggregation properties
        :param method: 'mean' or 'stddev'
        :param interval: interval in ms (will be stored in seconds)
        """
        if method not in self.list_agg_methods():
            raise KeyError(f'unsupported agg_method {method}.')
        self.aggregation = method
        if interval is not None:
            self.interval = int(interval) / 1000

    def list_agg_methods(self) -> list:
        """
        get a list of all authorized methods.
        :return: list of all authorized methods.
        """
        return [
            "mean", "stddev", "mode", "median", "count", "sum", "first", "last", "max", "min"
        ]

    def select_template(self,
                        template_id=None,
                        template_key=None,
                        twin_id=None,
                        twin_hardware_id=None):
        """
        Select a template and its registration.
        :param template_id: template UUID
        :param template_key: template key ( ignored if template_id specified )
        :param twin_id: Digital Twin UUID
        :param twin_hardware_id: hardware ID of Digital Twin ( ignored if twin_id specified )
        """
        if template_id is None and template_key is None and twin_id is None and twin_hardware_id is None:
            self.template = None
            return
        else:
            self.template = {}

            if template_id is not None:
                if isinstance(template_id, uuid.UUID):
                    self.template['template_id'] = template_id
                else:
                    self.template['template_id'] = uuid.UUID(template_id)
            elif template_key is not None:
                self.template['template_key'] = str(template_key)

            if twin_id is not None:
                if isinstance(twin_id, uuid.UUID):
                    self.template['twin_id'] = twin_id
                else:
                    self.template['twin_id'] = uuid.UUID(twin_id)

            elif twin_hardware_id is not None:
                self.template['twin_hardware_id'] = str(twin_hardware_id)

    def get_params(self):
        """
        get a list of all parameters.
        """
        params = []

        if self.start is not None and isinstance(self.start, str) and self.start.startswith("@"):
            params.append(self.start[1:])

        if self.end is not None and isinstance(self.end, str) and self.end.startswith("@"):
            params.append(self.end[1:])

        return list(set(params))

    def set_param(self, name: str, value):
        """
        set value of parameter based on his name.
        """
        assigned = False

        if value is None:
            raise ValueError(f'please provide a valid param value for {name}')

        if self.start is not None and isinstance(self.start, str) \
                and self.start.startswith("@") and self.start[1:] == name:
            self.start = value
            assigned = True

        if self.end is not None and isinstance(self.end, str) \
                and self.end.startswith("@") and self.end[1:] == name:
            self.end = value
            assigned = True

        if not assigned:
            raise KeyError(f'parameter {name} not found in request.')

    def to_json(self, target: str = None):
        """
        convert to a dict JSON compatible for all properties. For query only, use prepare().

        :return: a dict JSON compatible
        """

        # Prepare is 'to_json' without future obsolete properties
        obj = self.prepare()

        if self.target_feat is not None:
            obj["target_feat"] = {
                "sensor": self.target_feat["sensor"],
                "operator": self.target_feat["operator"],
                "threshold": self.target_feat["threshold"]
            }
        if self.on_off_sensor is not None and self.restart_time is not None:
            obj["restart_filter"] = {
                "on_off_sensor": self.on_off_sensor,
                "stop_restart_time": self.restart_time
            }

        if self.sensitivity is not None:
            obj["sensitivity"] = self.sensitivity

        if self.extra_data is not None:
            obj["extra_data"] = self.extra_data

        return obj

    def from_json(self, json_data):
        """
        load a request based on dict from a JSON file.

        :param json_data: JSON formatted dictionnary object representing a query.
        """
        if "id" in json_data.keys():
            self.request_id = uuid.UUID(json_data["id"])

        if "name" in json_data.keys():
            self.name = json_data["name"]

        found_dps = False
        if "equipments_list" in json_data.keys():
            self.equipments = json_data["equipments_list"]
            for equipment in self.equipments:
                if "datapoints" not in equipment.keys():
                    raise KeyError("No 'data points' have been provided for equipment with id '" +
                                   str(equipment["id"]) + "'")
                else:
                    found_dps = True
        elif "datapoints" in json_data.keys():
            equipment = {"datapoints": []}
            for datapoint in json_data["datapoints"]:
                equipment["datapoints"].append(datapoint)
            self.equipments.append(equipment)
            found_dps = True

        if "template" in json_data.keys():
            self.template = json_data["template"]
            found_dps = True

        if not found_dps:
            raise KeyError('no equipments_list datapoints, nor template specified inside the request.')

        if "timeframe" not in json_data.keys():
            raise KeyError("No 'time range' have been selected, please set it up and re-try.")

        if "start" not in json_data["timeframe"].keys():
            raise KeyError("No 'start time' have been selected, please set it up and re-try.")

        if isinstance(json_data["timeframe"]["start"], str):
            self.start = json_data["timeframe"]["start"]
        else:
            self.start = datetime.fromtimestamp(json_data["timeframe"]["start"] / 1000, timezone.utc)

        if "end" not in json_data["timeframe"].keys():
            raise KeyError("No 'end time' have been selected, please set it up and re-try.")

        if isinstance(json_data["timeframe"]["end"], str):
            self.end = json_data["timeframe"]["end"]
        else:
            self.end = datetime.fromtimestamp(json_data["timeframe"]["end"] / 1000, timezone.utc)

        if "aggregations" not in json_data.keys():
            raise KeyError("No 'aggregations' have been selected, please set it up and re-try.")

        if "agg_method" not in json_data["aggregations"].keys():
            raise KeyError("No 'Aggregation Method' have been selected, please set it up and re-try.")
        if json_data["aggregations"]["agg_method"] not in self.list_agg_methods():
            raise KeyError(f'unsupported agg_method {json_data["aggregations"]["agg_method"]}.')
        self.aggregation = json_data["aggregations"]["agg_method"]

        if "interval" not in json_data["aggregations"].keys():
            raise KeyError("No 'Aggregation Interval' have been selected, please set it up and re-try.")
        self.interval = int(json_data["aggregations"]["interval"] / 1000)

        if "filters" in json_data.keys():
            filters = {}
            for filter_data in json_data["filters"]:
                filter_reformat = {}
                for operator in json_data["filters"][filter_data]:
                    if operator in filter_map.keys():
                        filter_reformat[filter_map[operator]] = json_data["filters"][filter_data][operator]
                    elif operator not in filter_map.values():
                        raise ValueError(f'invalid request filter operator {operator}')
                    else:
                        filter_reformat[operator] = json_data["filters"][filter_data][operator]
                filters[filter_data] = filter_reformat
            self.filters = filters
        else:
            self.filters = {}

        if "options" in json_data.keys():
            self.options = json_data["options"]
        else:
            self.options = {}

        if "connections" in json_data.keys():
            self.connections = json_data["connections"]

        if "null" in json_data.keys():
            self.null = json_data["null"]

        if "target_feat" in json_data.keys():
            self.target_feat = json_data["target_feat"]
            if "sensor" not in self.target_feat.keys():
                raise KeyError("No 'sensor' have been declared inside the target feature, this is a technical error.")
            if "operator" not in self.target_feat.keys():
                raise KeyError("No 'operator' have been declared inside the target feature, this is a technical error.")
            if "threshold" not in self.target_feat.keys():
                raise KeyError("No 'threshold' have been declared inside the target feature, this is a technical error.")

        if "restart_filter" in json_data.keys():
            self.on_off_sensor = json_data["restart_filter"]["on_off_sensor"]
            self.restart_time = json_data["restart_filter"]["stop_restart_time"]

        if "sensitivity" in json_data.keys():
            self.sensitivity = json_data["sensitivity"]

        if "extra_data" in json_data.keys():
            self.extra_data = json_data["extra_data"]


